"""
This module provides a framework for managing recipients. It defines three class
`Recipient`, `RecipientManager` and `PythonRecipient`. It also defines a metaclass
`RecipientMeta` for the `Recipient` class.

`Recipient` is an abstract base class that defines the basic structure and methods
for a recipient. It provides an interface for processing a message and an
asynchronous processing method, along with a UUID generator for generating unique
message IDs.

`RecipientManager` is a class that manages a registry of recipients. It allows users
to register and unregister recipients by name, and provides an interface for
accessing and retrieving registered recipients. The class also provides a property
for listing all available recipients, along with their descriptions.

`PythonRecipient` is a class that provides an interface for processing Python code.
It provides an asynchronous context manager for executing code in a separate thread,
and an asynchronous processing method for processing messages.

`RecipientMeta` is a metaclass for the `Recipient` class. It sets the `RECIPIENT_NAME`
attribute of the class to the name of the class, if the attribute is not already set.
"""
import asyncio
import uuid
from abc import ABCMeta
from abc import abstractmethod
from typing import Callable
from typing import Type

from async_tio import Tio


class RecipientMeta(ABCMeta):
    """
    Metaclass for the Recipient class.
    """

    def __new__(cls, name, bases, namespace, /, **kwargs):
        if "RECIPIENT_NAME" not in namespace or not namespace["RECIPIENT_NAME"]:
            namespace["RECIPIENT_NAME"] = name
        return super().__new__(cls, name, bases, namespace, **kwargs)


class Recipient(metaclass=RecipientMeta):
    """
    Recipient base class.

    Attributes
    ----------
    DESCRIPTION : str
        Description of the recipient for humans.
    API_DOCS : str
        Documentation of the recipient for machines.
    REQUIRED_ARGS : list
        List of required parameters.
    EXAMPLE_MESSAGES : list
        List of example messages.
    RECIPIENT_NAME : str
        Name of the recipient. Defaults to the class name.

    Methods
    -------
    process(message: dict, **kwargs: dict)
        Process a message.
    aprocess(message: dict, **kwargs: dict)
        Asynchronously process a message.
    """

    DESCRIPTION = ""
    API_DOCS = ""
    REQUIRED_ARGS = []
    EXAMPLE_MESSAGES = []
    RECIPIENT_NAME = ""
    # Defaults to the class name. This process is handled by the metaclass.

    @abstractmethod
    def process(self, message: dict, **kwargs: dict) -> dict:
        """
        Process a message.

        Parameters
        ----------
        message : dict
            The message to process.
        kwargs : dict
            Additional arguments.

        Returns
        -------
        dict
            The message which need to be sent.
        """
        raise NotImplementedError("Method needs to be implemented in subclass.")

    @abstractmethod
    async def aprocess(self, message: dict, **kwargs: dict) -> dict:
        """
        Asynchronously process a message.

        Parameters
        ----------
        message : dict
            The message to process.
        kwargs : dict
            Additional arguments.

        Returns
        -------
        dict
            The message which need to be sent.
        """
        raise NotImplementedError("Method needs to be implemented in subclass.")

    def _uuid(self) -> str:
        """
        Generate a UUID.

        Returns
        -------
        str
            A UUID.
        """
        return str(uuid.uuid4())


class RecipientManager:
    """Recipient Manager class.

    Methods
    -------
    register(name: str) -> Callable[[Type[Recipient]], Type[Recipient]]
        Decorator for registering a recipient to the manager.
    __getitem__(name: str) -> Type[Recipient]
        Get a recipient from the manager.
    __setitem__(name: str, recipient_class: Type[Recipient])
        Register a recipient to the manager.
    __delitem__(name: str)
        Unregister a recipient from the manager.

    Properties
    ----------
    available_recipients : dict[str, str]
        Dictionary containing the available recipients.
    """

    def __init__(self):
        self.__registry_list = {}

    def register(self, name="") -> Callable[[Type[Recipient]], Type[Recipient]]:
        """Decorator for registering a recipient to the manager.

        Parameters
        ----------
        name : str
            The name of the recipient to register.

        Returns
        -------
        Callable[[Type[Recipient]], Type[Recipient]]
            Decorator function for registering the recipient.
        """

        def decorator(recipient_class: Type[Recipient]) -> Type[Recipient]:
            self[name or recipient_class.RECIPIENT_NAME] = recipient_class
            return recipient_class

        return decorator

    @property
    def available_recipients(self) -> dict[str, str]:
        """Dictionary containing the available recipients.

        Returns
        -------
        dict[str, str]
            A dictionary where the keys are the names of the available recipients, and
            the values are their descriptions.
        """
        return {
            name: recipient_class.DESCRIPTION
            for name, recipient_class in self.__registry_list.items()
        }

    def __getitem__(self, name: str) -> Type[Recipient]:
        """Get a recipient from the manager.

        Parameters
        ----------
        name : str
            The name of the recipient to get.

        Returns
        -------
        Type[Recipient]
            The recipient class.

        Raises
        ------
        KeyError
            Raised if the recipient is not registered.
        """
        if name not in self.__registry_list:
            raise KeyError(f"Recipient '{name}' is not registered.")
        return self.__registry_list[name]

    def __setitem__(self, name: str, recipient_class: Type[Recipient]):
        """Register a recipient to the manager.

        Parameters
        ----------
        name : str
            The name of the recipient to set.
        recipient_class : Type[Recipient]
            The class of the recipient to set.

        Raises
        ------
        KeyError
            Raised if the recipient is already registered.
        """
        if name in self.__registry_list:
            raise KeyError(f"Recipient '{name}' is already registered.")
        self.__registry_list[name] = recipient_class

    def __delitem__(self, name: str):
        """Unregister a recipient from the manager.

        Parameters
        ----------
        name : str
            The name of the recipient to unregister.
        """
        del self.__registry_list[name]

    def __contains__(self, name: str) -> bool:
        """Check if a recipient is registered.

        Parameters
        ----------
        name : str
            The name of the recipient to check.

        Returns
        -------
        bool
            True if the recipient is registered, False otherwise.
        """
        return name in self.__registry_list


class PythonRecipient(Recipient):
    """
    Python recipient class.
    """

    DESCRIPTION = "Python Interpreter Remote"
    API_DOCS = (
        "You have the tool python, and you can run any python code because it run in a sandbox.\n"
        "But you must send the code to recipient 'python' to run it."
    )
    RECIPIENT_NAME = "python"
    # Because OpenAI has finetuned GPT-3.5 to use python, we don't need examples.

    def __init__(self) -> None:
        super().__init__()
        self.tio = Tio()

    async def aprocess(self, message: dict, **kwargs: dict):
        result = await self.tio.execute(message["message"], language="python3")
        result = result.stdout[:-1]
        return {
            "id": self._uuid(),
            "author": {
                "role": "tool",
                "name": self.RECIPIENT_NAME,
            },
            "content": {
                "content_type": "text",
                "parts": [f"```output\n{result}\n```"],
            },
        }

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.tio.close()

    def process(self, message: dict, **kwargs: dict) -> dict:
        return asyncio.get_event_loop().run_until_complete(
            self.aprocess(message, **kwargs),
        )



import asyncio
import datetime

import revChatGPT.V1
from revChatGPT.recipient import PythonRecipient
from revChatGPT.V1 import Chatbot


def test_recipient() -> None:
    """
    Test the chatbot
    """
    config = revChatGPT.V1.configure()
    assert "access_token" in config
    cbt = Chatbot(config)
    assert cbt is not None

    # Try to get the recipient which is not registered
    try:
        _ = cbt.recipients["python"]
        raise AssertionError
    except Exception as ex:
        assert isinstance(ex, KeyError)
        assert ex.args[0] == "Recipient 'python' is not registered."

    # Register the recipient
    cbt.recipients["python"] = PythonRecipient
    assert cbt.recipients["python"].DESCRIPTION == "Python Interpreter Remote"

    # Try to register the recipient which is already registered
    try:
        cbt.recipients["python"] = PythonRecipient
        raise AssertionError
    except Exception as ex:
        assert isinstance(ex, KeyError)
        assert ex.args[0] == "Recipient 'python' is already registered."

    # To get the recipient instance
    python = cbt.recipients["python"]()

    api_docs = f"""

    Knowledge cutoff: 2021-09
    Current date: {datetime.datetime.now().strftime("%Y-%m-%d")}

    ###Available Tools:
    python

    {python.API_DOCS}
    """
    for _ in cbt.ask(f"You are ChatGPT.{api_docs}"):
        pass

    # Test the python recipient
    result = {}
    for _ in cbt.ask("Please calculate 27! + 8! by using python."):
        result = _
    assert result
    assert not result["end_turn"]

    times = 0
    while not result.get("end_turn", True):
        times += 1
        recipient_name = result["recipient"]
        if times >= 3:
            break
        assert recipient_name == "python"
        msg = asyncio.get_event_loop().run_until_complete(
            python.aprocess(message=result.copy()),
        )
        assert msg["content"]["parts"][0].startswith("```")
        for _ in cbt.post_messages([msg]):
            result = _

    assert times == 1
    assert "10888869450418352160768040320" in result["message"].replace(",", "")



from pathlib import Path

from setuptools import find_namespace_packages
from setuptools import setup

DOCS_PATH = Path(__file__).parents[0] / "docs/README.md"
PATH = Path("README.md")
if not PATH.exists():
    with open(DOCS_PATH, encoding="utf-8") as f1:
        with open(PATH, "w+", encoding="utf-8") as f2:
            f2.write(f1.read())

setup(
    name="revChatGPT",
    version="5.1.1",
    description="ChatGPT is a reverse engineering of OpenAI's ChatGPT API",
    long_description=open(PATH, encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/acheong08/ChatGPT",
    project_urls={
        "Bug Report": "https://github.com/acheong08/ChatGPT/issues/new?assignees=&labels=bug-report&template=bug_report.yml&title=%5BBug%5D%3A+",
        "Feature request": "https://github.com/acheong08/ChatGPT/issues/new?assignees=&labels=enhancement&template=feature_request.yml&title=%5BFeature+Request%5D%3A+",
    },
    author="Antonio Cheong",
    author_email="acheong@student.dalat.org",
    license="GNU General Public License v2.0",
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    py_modules=["revChatGPT"],
    package_data={"": ["*.json"]},
    install_requires=[
        "OpenAIAuth==0.3.6",
        "requests[socks]",
        "httpx[socks]",
        "async_tio",
        "prompt-toolkit",
        "tiktoken>=0.3.0",
        "openai",
    ],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



"""
Main CLI
"""
import argparse
import sys

from . import __version__
from . import typings as t
from . import V1
from . import V3

__all__ = ()


def main():
    """
    main function for CLI
    """
    parser = argparse.ArgumentParser(
        description="ChatGPT - A command-line interface to OpenAI's ChatGPT (https://chat.openai.com/chat)",
        epilog="Repo: github.com/acheong08/ChatGPT",
        allow_abbrev=True,
    )
    parser.add_argument(
        "--V1",
        action="store_true",
        help="Use the website version of ChatGPT",
    )
    parser.add_argument(
        "--V3",
        action="store_true",
        help="Use the API version of ChatGPT",
    )
    args, _ = parser.parse_known_args()
    mode = "V1" if args.V1 else "V3" if args.V3 else input("Version (V1/V3):")

    if mode == "V1":
        print(
            f"""
        ChatGPT - A command-line interface to OpenAI's ChatGPT (https://chat.openai.com/chat)
        Repo: github.com/acheong08/ChatGPT
        Version: {__version__}
        """,
        )
        print("Type '!help' to show a full list of commands")
        print(
            f"{V1.bcolors.BOLD}{V1.bcolors.WARNING}Press Esc followed by Enter or Alt+Enter to send a message.{V1.bcolors.ENDC}",
        )
        V1.main(V1.configure())
    elif mode == "V3":
        try:
            V3.main()
        except Exception as exc:
            error = t.CLIError("Command line program unknown error")
            raise error from exc
        except KeyboardInterrupt:
            print("\nExiting...")
            sys.exit()
    else:
        error = NotImplementedError(f"Unknown version: {mode}")
        raise error


if __name__ == "__main__":
    main()
